@@ -18,6 +18,8 @@
 #include <QGuiApplication>
 #include <QtGui/private/qwindow_p.h>
 
+#include <QtCore/qpointer.h>
+
 // -------------------------------------------------------------------------
 
 static QUIView *focusView()
@@ -119,7 +121,7 @@ static QUIView *focusView()
 {
     [self keyboardWillOrDidChange:notification];
 
-    UIResponder *firstResponder = [UIResponder currentFirstResponder];
+    UIResponder *firstResponder = [UIResponder qt_currentFirstResponder];
     if (![firstResponder isKindOfClass:[QIOSTextInputResponder class]])
         return;
 
@@ -174,7 +176,11 @@ static QUIView *focusView()
 {
     [super touchesBegan:touches withEvent:event];
 
-    Q_ASSERT(m_context->isInputPanelVisible());
+    if (!m_context->isInputPanelVisible()) {
+        qImDebug("keyboard was hidden by sliding it down, disabling hide-keyboard gesture");
+        self.enabled = NO;
+        return;
+    }
 
     if ([touches count] != 1)
         self.state = UIGestureRecognizerStateFailed;
@@ -228,7 +234,7 @@ static QUIView *focusView()
 
     if (self.state == UIGestureRecognizerStateBegan) {
         qImDebug("hide keyboard gesture was triggered");
-        UIResponder *firstResponder = [UIResponder currentFirstResponder];
+        UIResponder *firstResponder = [UIResponder qt_currentFirstResponder];
         Q_ASSERT([firstResponder isKindOfClass:[QIOSTextInputResponder class]]);
         [firstResponder resignFirstResponder];
     }
@@ -297,11 +303,7 @@ QIOSInputContext::QIOSInputContext()
     , m_keyboardHideGesture([[QIOSKeyboardListener alloc] initWithQIOSInputContext:this])
     , m_textResponder(0)
 {
-    if (isQtApplication()) {
-        QIOSScreen *iosScreen = static_cast<QIOSScreen*>(QGuiApplication::primaryScreen()->handle());
-        [iosScreen->uiWindow() addGestureRecognizer:m_keyboardHideGesture];
-    }
-
+    Q_ASSERT(!qGuiApp->focusWindow());
     connect(qGuiApp, &QGuiApplication::focusWindowChanged, this, &QIOSInputContext::focusWindowChanged);
 }
 
@@ -346,7 +348,7 @@ void QIOSInputContext::clearCurrentFocusObject()
 
 void QIOSInputContext::updateKeyboardState(NSNotification *notification)
 {
-#ifdef Q_OS_TVOS
+#if defined(Q_OS_TVOS) || defined(Q_OS_VISIONOS)
     Q_UNUSED(notification);
 #else
     static CGRect currentKeyboardRect = CGRectZero;
@@ -369,7 +371,8 @@ void QIOSInputContext::updateKeyboardState(NSNotification *notification)
         // with input-accessory-views. The reason for using frameEnd here (the future state),
         // instead of the current state reflected in frameBegin, is that QInputMethod::isVisible()
         // is documented to reflect the future state in the case of animated transitions.
-        m_keyboardState.keyboardVisible = CGRectIntersectsRect(frameEnd, [UIScreen mainScreen].bounds);
+        m_keyboardState.keyboardVisible = !CGRectIsEmpty(UIScreen.mainScreen.bounds) &&
+            !CGRectIsEmpty(frameEnd) && CGRectIntersectsRect(frameEnd, UIScreen.mainScreen.bounds);
 
         // Used for auto-scroller, and will be used for animation-signal in the future
         m_keyboardState.keyboardEndRect = frameEnd;
@@ -436,6 +439,7 @@ UIView *QIOSInputContext::scrollableRootView()
 
 void QIOSInputContext::scrollToCursor()
 {
+#if !defined(Q_OS_VISIONOS)
     if (!isQtApplication())
         return;
 
@@ -492,6 +496,7 @@ void QIOSInputContext::scrollToCursor()
     } else {
         scroll(0);
     }
+#endif
 }
 
 void QIOSInputContext::scroll(int y)
@@ -603,12 +608,15 @@ void QIOSInputContext::setFocusObject(QObject *focusObject)
 
 void QIOSInputContext::focusWindowChanged(QWindow *focusWindow)
 {
-    Q_UNUSED(focusWindow);
-
     qImDebug() << "new focus window =" << focusWindow;
 
     reset();
 
+    if (isQtApplication()) {
+        [m_keyboardHideGesture.view removeGestureRecognizer:m_keyboardHideGesture];
+        [focusView().window addGestureRecognizer:m_keyboardHideGesture];
+    }
+
     // The keyboard rectangle depend on the focus window, so
     // we need to re-evaluate the keyboard state.
     updateKeyboardState();
@@ -632,8 +640,10 @@ void QIOSInputContext::update(Qt::InputMethodQueries updatedProperties)
     // focus object. We try to detect code paths that fail this assertion and smooth
     // over the situation by doing a manual update of the focus object.
     if (qApp->focusObject() != m_imeState.focusObject && updatedProperties != Qt::ImQueryAll) {
-        qWarning() << "stale focus object" << static_cast<void *>(m_imeState.focusObject)
-                   << ", doing manual update";
+        qCWarning(lcQpaInputMethods).verbosity(0) << "Updating input context" << updatedProperties
+            << "with last reported focus object" << m_imeState.focusObject
+            << "but qGuiApp reports" << qApp->focusObject()
+            << "which means someone failed to call QPlatformInputContext::setFocusObject()";
         setFocusObject(qApp->focusObject());
         return;
     }
