@@ -1,30 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the test suite of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QTest>
 #include <QSignalSpy>
@@ -64,9 +39,12 @@ private slots:
     void classNameFirstInStringData();
 
     void propertyMetaType();
+    void enumCloning();
 
     void cleanupTestCase();
 
+    void ownMetaTypeNoProperties();
+
 private:
     static bool checkForSideEffects
         (const QMetaObjectBuilder& builder,
@@ -97,7 +75,9 @@ class SomethingOfEverything : public QObject
     Q_PROPERTY(SomethingFlagEnum fprop READ fprop)
     Q_PROPERTY(QLocale::Language language READ language)
     Q_ENUMS(SomethingEnum)
-    Q_FLAGS(SomethingFlagEnum)
+    Q_FLAGS(SomethingFlag)
+    Q_ENUMS(SomethingEnum64)
+    Q_FLAGS(SomethingFlag64)
 public:
     Q_INVOKABLE SomethingOfEverything() {}
     ~SomethingOfEverything() {}
@@ -108,11 +88,25 @@ public:
         JKL = 10
     };
 
+    enum SomethingEnum64 : qint64
+    {
+        MNO = -1,
+        PQR = 0x1'2345'5678,
+    };
+
     enum SomethingFlagEnum
     {
         XYZ = 1,
         UVW = 8
     };
+    Q_DECLARE_FLAGS(SomethingFlag, SomethingFlagEnum)
+
+    enum SomethingFlagEnum64 : quint64
+    {
+        RST = Q_UINT64_C(1) << 31,
+        OPQ = Q_UINT64_C(1) << 63,
+    };
+    Q_DECLARE_FLAGS(SomethingFlag64, SomethingFlagEnum64)
 
     Q_INVOKABLE Q_SCRIPTABLE void method1() {}
 
@@ -264,7 +258,7 @@ void tst_QMetaObjectBuilder::method()
     method1.setParameterNames(QList<QByteArray>() << "a" << "b");
     method1.setTag("tag");
     method1.setAccess(QMetaMethod::Private);
-    method1.setAttributes(42);
+    method1.setAttributes(QMetaMethod::Cloned);
     method1.setRevision(123);
 
     // Check that method1 is changed, but method2 is not.
@@ -275,7 +269,7 @@ void tst_QMetaObjectBuilder::method()
     QCOMPARE(method1.parameterNames(), QList<QByteArray>() << "a" << "b");
     QCOMPARE(method1.tag(), QByteArray("tag"));
     QCOMPARE(method1.access(), QMetaMethod::Private);
-    QCOMPARE(method1.attributes(), 42);
+    QCOMPARE(method1.attributes(), QMetaMethod::Cloned);
     QCOMPARE(method1.revision(), 123);
     QCOMPARE(method1.index(), 0);
     QCOMPARE(method2.signature(), QByteArray("bar(QString)"));
@@ -295,7 +289,7 @@ void tst_QMetaObjectBuilder::method()
     method2.setParameterNames(QList<QByteArray>() << "c");
     method2.setTag("Q_FOO");
     method2.setAccess(QMetaMethod::Protected);
-    method2.setAttributes(24);
+    method2.setAttributes(QMetaMethod::Scriptable);
     method2.setRevision(321);
 
     // This time check that only method2 changed.
@@ -306,7 +300,7 @@ void tst_QMetaObjectBuilder::method()
     QCOMPARE(method1.parameterNames(), QList<QByteArray>() << "a" << "b");
     QCOMPARE(method1.tag(), QByteArray("tag"));
     QCOMPARE(method1.access(), QMetaMethod::Private);
-    QCOMPARE(method1.attributes(), 42);
+    QCOMPARE(method1.attributes(), QMetaMethod::Cloned);
     QCOMPARE(method1.revision(), 123);
     QCOMPARE(method1.index(), 0);
     QCOMPARE(method2.signature(), QByteArray("bar(QString)"));
@@ -316,7 +310,7 @@ void tst_QMetaObjectBuilder::method()
     QCOMPARE(method2.parameterNames(), QList<QByteArray>() << "c");
     QCOMPARE(method2.tag(), QByteArray("Q_FOO"));
     QCOMPARE(method2.access(), QMetaMethod::Protected);
-    QCOMPARE(method2.attributes(), 24);
+    QCOMPARE(method2.attributes(), QMetaMethod::Scriptable);
     QCOMPARE(method2.revision(), 321);
     QCOMPARE(method2.index(), 1);
     QCOMPARE(builder.methodCount(), 2);
@@ -332,7 +326,7 @@ void tst_QMetaObjectBuilder::method()
     QCOMPARE(method2.parameterNames(), QList<QByteArray>() << "c");
     QCOMPARE(method2.tag(), QByteArray("Q_FOO"));
     QCOMPARE(method2.access(), QMetaMethod::Protected);
-    QCOMPARE(method2.attributes(), 24);
+    QCOMPARE(method2.attributes(), QMetaMethod::Scriptable);
     QCOMPARE(method2.revision(), 321);
     QCOMPARE(method2.index(), 0);
 
@@ -467,7 +461,7 @@ void tst_QMetaObjectBuilder::constructor()
     ctor1.setParameterNames(QList<QByteArray>() << "a" << "b");
     ctor1.setTag("tag");
     ctor1.setAccess(QMetaMethod::Private);
-    ctor1.setAttributes(42);
+    ctor1.setAttributes(QMetaMethod::Scriptable);
 
     // Check that ctor1 is changed, but ctor2 is not.
     QCOMPARE(ctor1.signature(), QByteArray("foo(QString,int)"));
@@ -477,7 +471,7 @@ void tst_QMetaObjectBuilder::constructor()
     QCOMPARE(ctor1.parameterNames(), QList<QByteArray>() << "a" << "b");
     QCOMPARE(ctor1.tag(), QByteArray("tag"));
     QCOMPARE(ctor1.access(), QMetaMethod::Private);
-    QCOMPARE(ctor1.attributes(), 42);
+    QCOMPARE(ctor1.attributes(), QMetaMethod::Scriptable);
     QCOMPARE(ctor1.index(), 0);
     QCOMPARE(ctor2.signature(), QByteArray("bar(QString)"));
     QCOMPARE(ctor2.methodType(), QMetaMethod::Constructor);
@@ -495,7 +489,7 @@ void tst_QMetaObjectBuilder::constructor()
     ctor2.setParameterNames(QList<QByteArray>() << "c");
     ctor2.setTag("Q_FOO");
     ctor2.setAccess(QMetaMethod::Protected);
-    ctor2.setAttributes(24);
+    ctor2.setAttributes(QMetaMethod::Compatibility);
 
     // This time check that only ctor2 changed.
     QCOMPARE(ctor1.signature(), QByteArray("foo(QString,int)"));
@@ -505,7 +499,7 @@ void tst_QMetaObjectBuilder::constructor()
     QCOMPARE(ctor1.parameterNames(), QList<QByteArray>() << "a" << "b");
     QCOMPARE(ctor1.tag(), QByteArray("tag"));
     QCOMPARE(ctor1.access(), QMetaMethod::Private);
-    QCOMPARE(ctor1.attributes(), 42);
+    QCOMPARE(ctor1.attributes(), QMetaMethod::Scriptable);
     QCOMPARE(ctor1.index(), 0);
     QCOMPARE(ctor2.signature(), QByteArray("bar(QString)"));
     QCOMPARE(ctor2.methodType(), QMetaMethod::Constructor);
@@ -514,7 +508,7 @@ void tst_QMetaObjectBuilder::constructor()
     QCOMPARE(ctor2.parameterNames(), QList<QByteArray>() << "c");
     QCOMPARE(ctor2.tag(), QByteArray("Q_FOO"));
     QCOMPARE(ctor2.access(), QMetaMethod::Protected);
-    QCOMPARE(ctor2.attributes(), 24);
+    QCOMPARE(ctor2.attributes(), QMetaMethod::Compatibility);
     QCOMPARE(ctor2.index(), 1);
     QCOMPARE(builder.constructorCount(), 2);
 
@@ -529,7 +523,7 @@ void tst_QMetaObjectBuilder::constructor()
     QCOMPARE(ctor2.parameterNames(), QList<QByteArray>() << "c");
     QCOMPARE(ctor2.tag(), QByteArray("Q_FOO"));
     QCOMPARE(ctor2.access(), QMetaMethod::Protected);
-    QCOMPARE(ctor2.attributes(), 24);
+    QCOMPARE(ctor2.attributes(), QMetaMethod::Compatibility);
     QCOMPARE(ctor2.index(), 0);
 
     // Perform index-based lookup again.
@@ -801,6 +795,26 @@ void tst_QMetaObjectBuilder::notifySignal()
 
 void tst_QMetaObjectBuilder::enumerator()
 {
+    static const QtPrivate::QMetaTypeInterface fooFlagMetaType = {
+        0,
+        8,
+        8,
+        QMetaType::IsEnumeration | QMetaType::IsUnsignedEnumeration | QMetaType::RelocatableType,
+        {},
+        nullptr,
+        "fooFlag",
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+        nullptr,
+    };
+
     QMetaObjectBuilder builder;
 
     // Add an enumerator and check its attributes.
@@ -808,6 +822,7 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum1.name(), QByteArray("foo"));
     QVERIFY(!enum1.isFlag());
     QVERIFY(!enum1.isScoped());
+    QVERIFY(!enum1.is64Bit());
     QCOMPARE(enum1.keyCount(), 0);
     QCOMPARE(enum1.index(), 0);
     QCOMPARE(builder.enumeratorCount(), 1);
@@ -817,6 +832,7 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum2.name(), QByteArray("bar"));
     QVERIFY(!enum2.isFlag());
     QVERIFY(!enum2.isScoped());
+    QVERIFY(!enum2.is64Bit());
     QCOMPARE(enum2.keyCount(), 0);
     QCOMPARE(enum2.index(), 1);
     QCOMPARE(builder.enumeratorCount(), 2);
@@ -832,15 +848,20 @@ void tst_QMetaObjectBuilder::enumerator()
     enum1.setIsFlag(true);
     enum1.setIsScoped(true);
     enum1.setEnumName(QByteArrayLiteral("fooFlag"));
+    enum1.setMetaType(QMetaType(&fooFlagMetaType));
+    QVERIFY(enum1.is64Bit());
+    enum1.setIs64Bit(false);
+    QVERIFY(!enum1.is64Bit());
     QCOMPARE(enum1.addKey("ABC", 0), 0);
     QCOMPARE(enum1.addKey("DEF", 1), 1);
-    QCOMPARE(enum1.addKey("GHI", -1), 2);
+    QCOMPARE(enum1.addKey("GHI", -2), 2);
 
     // Check that enum1 is changed, but enum2 is not.
     QCOMPARE(enum1.name(), QByteArray("foo"));
     QVERIFY(enum1.isFlag());
     QVERIFY(enum1.isScoped());
     QCOMPARE(enum1.enumName(), QByteArray("fooFlag"));
+    QCOMPARE(enum1.metaType(), QMetaType(&fooFlagMetaType));
     QCOMPARE(enum1.keyCount(), 3);
     QCOMPARE(enum1.index(), 0);
     QCOMPARE(enum1.key(0), QByteArray("ABC"));
@@ -849,7 +870,10 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum1.key(3), QByteArray());
     QCOMPARE(enum1.value(0), 0);
     QCOMPARE(enum1.value(1), 1);
-    QCOMPARE(enum1.value(2), -1);
+    QCOMPARE(enum1.value(2), -2);
+    QCOMPARE(enum1.value64(0), 0);
+    QCOMPARE(enum1.value64(1), 1);
+    QCOMPARE(enum1.value64(2), uint(-2));
     QCOMPARE(enum2.name(), QByteArray("bar"));
     QVERIFY(!enum2.isFlag());
     QVERIFY(!enum2.isScoped());
@@ -859,12 +883,14 @@ void tst_QMetaObjectBuilder::enumerator()
     // Modify the attributes on enum2.
     enum2.setIsFlag(true);
     QCOMPARE(enum2.addKey("XYZ", 10), 0);
-    QCOMPARE(enum2.addKey("UVW", 19), 1);
+    QCOMPARE(enum2.addKey("UVW", quint64(1) << 32), 1);
+    QVERIFY(enum2.is64Bit());
 
     // This time check that only method2 changed.
     QCOMPARE(enum1.name(), QByteArray("foo"));
     QVERIFY(enum1.isFlag());
     QVERIFY(enum1.isScoped());
+    QVERIFY(!enum1.is64Bit());
     QCOMPARE(enum1.keyCount(), 3);
     QCOMPARE(enum1.index(), 0);
     QCOMPARE(enum1.key(0), QByteArray("ABC"));
@@ -873,17 +899,35 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum1.key(3), QByteArray());
     QCOMPARE(enum1.value(0), 0);
     QCOMPARE(enum1.value(1), 1);
-    QCOMPARE(enum1.value(2), -1);
+    QCOMPARE(enum1.value(2), -2);
+    QCOMPARE(enum1.value64(0), 0);
+    QCOMPARE(enum1.value64(1), 1);
+    QCOMPARE(enum1.value64(2), uint(-2));
     QCOMPARE(enum2.name(), QByteArray("bar"));
     QVERIFY(enum2.isFlag());
     QVERIFY(!enum2.isScoped());
+    QVERIFY(enum2.is64Bit());
     QCOMPARE(enum2.keyCount(), 2);
     QCOMPARE(enum2.index(), 1);
     QCOMPARE(enum2.key(0), QByteArray("XYZ"));
     QCOMPARE(enum2.key(1), QByteArray("UVW"));
     QCOMPARE(enum2.key(2), QByteArray());
     QCOMPARE(enum2.value(0), 10);
-    QCOMPARE(enum2.value(1), 19);
+    QCOMPARE(enum2.value(1), 0);    // truncated!
+    QCOMPARE(enum2.value64(0), 10);
+    QCOMPARE(enum2.value64(1), quint64(1) << 32);
+
+    // Reset enum2 to 32 bits
+    enum2.setIs64Bit(false);
+    QCOMPARE(enum2.value(0), 10);
+    QCOMPARE(enum2.value(1), 0);
+    QCOMPARE(enum2.value64(0), 10);
+    QCOMPARE(enum2.value64(1), 0);
+
+    // Reset back restores it
+    enum2.setIs64Bit(true);
+    QCOMPARE(enum2.value64(0), 10);
+    QCOMPARE(enum2.value64(1), quint64(1) << 32);
 
     // Remove enum1 key
     enum1.removeKey(2);
@@ -898,6 +942,9 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum1.value(0), 0);
     QCOMPARE(enum1.value(1), 1);
     QCOMPARE(enum1.value(2), -1);
+    QCOMPARE(enum1.value64(0), 0);
+    QCOMPARE(enum1.value64(1), 1);
+    QCOMPARE(enum1.value64(2), std::nullopt);
     QCOMPARE(enum2.name(), QByteArray("bar"));
     QVERIFY(enum2.isFlag());
     QVERIFY(!enum2.isScoped());
@@ -907,7 +954,9 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum2.key(1), QByteArray("UVW"));
     QCOMPARE(enum2.key(2), QByteArray());
     QCOMPARE(enum2.value(0), 10);
-    QCOMPARE(enum2.value(1), 19);
+    QCOMPARE(enum2.value(1), 0);        // truncated!
+    QCOMPARE(enum2.value64(0), 10);
+    QCOMPARE(enum2.value64(1), quint64(1) << 32);
 
     // Remove enum1 and check that enum2 becomes index 0.
     builder.removeEnumerator(0);
@@ -922,7 +971,9 @@ void tst_QMetaObjectBuilder::enumerator()
     QCOMPARE(enum2.key(1), QByteArray("UVW"));
     QCOMPARE(enum2.key(2), QByteArray());
     QCOMPARE(enum2.value(0), 10);
-    QCOMPARE(enum2.value(1), 19);
+    QCOMPARE(enum2.value(1), 0);        // truncated!
+    QCOMPARE(enum2.value64(0), 10);
+    QCOMPARE(enum2.value64(1), quint64(1) << 32);
 
     // Perform index-based lookup again.
     QCOMPARE(builder.indexOfEnumerator("foo"), -1);
@@ -1185,13 +1236,19 @@ static bool sameEnumerator(const QMetaEnum& enum1, const QMetaEnum& enum2)
     for (int index = 0; index < enum1.keyCount(); ++index) {
         if (QByteArray(enum1.key(index)) != QByteArray(enum2.key(index)))
             return false;
-        if (enum1.value(index) != enum2.value(index))
+        if (enum1.value64(index) != enum2.value64(index))
             return false;
     }
 
     if (QByteArray(enum1.scope()) != QByteArray(enum2.scope()))
         return false;
 
+    if (enum1.isScoped() != enum2.isScoped())
+        return false;
+
+    if (enum1.is64Bit() != enum2.is64Bit())
+        return false;
+
     return true;
 }
 
@@ -1481,9 +1538,7 @@ int TestObject::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
         if (_id < ownMethodCount)
             qt_static_metacall(this, _c, _id, _a);
         _id -= ownMethodCount;
-    }
-#ifndef QT_NO_PROPERTIES
-      else if (_c == QMetaObject::ReadProperty) {
+    } else if (_c == QMetaObject::ReadProperty) {
         void *_v = _a[0];
         switch (_id) {
         case 0: *reinterpret_cast< int*>(_v) = intProp(); break;
@@ -1506,7 +1561,6 @@ int TestObject::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
     } else if (_c == QMetaObject::ResetProperty) {
         _id -= ownPropertyCount;
     }
-#endif // QT_NO_PROPERTIES
     return _id;
 }
 
@@ -1524,8 +1578,8 @@ void tst_QMetaObjectBuilder::usage_signal()
 
     QSignalSpy propChangedSpy(testObject.data(), &TestObject::intPropChanged);
     testObject->emitIntPropChanged();
-    QCOMPARE(propChangedSpy.count(), 1);
-    QCOMPARE(propChangedSpy.at(0).count(), 1);
+    QCOMPARE(propChangedSpy.size(), 1);
+    QCOMPARE(propChangedSpy.at(0).size(), 1);
     QCOMPARE(propChangedSpy.at(0).at(0).toInt(), testObject->intProp());
 }
 
@@ -1540,7 +1594,7 @@ void tst_QMetaObjectBuilder::usage_property()
     QSignalSpy propChangedSpy(testObject.data(), &TestObject::intPropChanged);
     QVERIFY(testObject->intProp() != 123);
     testObject->setProperty("intProp", 123);
-    QCOMPARE(propChangedSpy.count(), 1);
+    QCOMPARE(propChangedSpy.size(), 1);
     prop = testObject->property("intProp");
     QCOMPARE(prop.metaType(), QMetaType(QMetaType::Int));
     QCOMPARE(prop.toInt(), 123);
@@ -1660,6 +1714,36 @@ void tst_QMetaObjectBuilder::propertyMetaType()
     free(mo);
 }
 
+void tst_QMetaObjectBuilder::enumCloning()
+{
+    QMetaObjectBuilder builder(&SomethingOfEverything::staticMetaObject);
+    auto smo = SomethingOfEverything::staticMetaObject;
+    auto mo = builder.toMetaObject();
+    auto cleanup = qScopeGuard([&]() { free(mo); });
+    QCOMPARE_EQ(mo->enumeratorCount(), smo.enumeratorCount());
+    for (int enumIndex = 0; enumIndex <  smo.enumeratorCount(); ++enumIndex) {
+        QMetaEnum metaEnumFromBuilder = mo->enumerator(enumIndex);
+        QMetaEnum originalMetaEnum = smo.enumerator(enumIndex);
+        QCOMPARE_EQ(metaEnumFromBuilder.metaType(), originalMetaEnum.metaType());
+        QCOMPARE_EQ(metaEnumFromBuilder.keyCount(), originalMetaEnum.keyCount());
+        for (int k = 0; k < originalMetaEnum.keyCount(); ++k) {
+            QCOMPARE_EQ(QByteArrayView(metaEnumFromBuilder.key(k)), QByteArrayView(originalMetaEnum.key(k)));
+            QCOMPARE_EQ(metaEnumFromBuilder.value(k), originalMetaEnum.value(k));
+            QCOMPARE_EQ(metaEnumFromBuilder.value64(k), originalMetaEnum.value64(k));
+        }
+    }
+}
+
+void tst_QMetaObjectBuilder::ownMetaTypeNoProperties()
+{
+    QMetaObjectBuilder builder;
+    builder.setClassName("NoProperties");
+    auto mo = builder.toMetaObject();
+    auto cleanup = qScopeGuard([&](){ free(mo); });
+    // own metatype should be invalid, as the dynamic metaobject has not been registered
+    QVERIFY(!mo->metaType().isValid());// should not crash
+}
+
 void tst_QMetaObjectBuilder::cleanupTestCase()
 {
     for (QMetaObject *obj: dynamicMetaObjectsPendingFree)
